/*
 * WALTZ
 *
 * The module tzaarlib starts the search according to the chosen algorithm and
 * does the control of the time limit via the Iterative Deepening for the
 * Alpha-beta based algorithms and for DFPNS via the estimation of the maximal
 * number of nodes that can be searched.
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */

#include "ai/lib.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <stdint.h>

const int BranchingFactorByFreeFields[] = // Averages from ~5000 positions from BAJ. These numbers contains duplicate positions.
	{9508, 9204, 9000, 8256, 8000, 7428, 7000, 6839, 6400, 5980, 5700, 5430, 5000, 4779, 4400, 4262, 4000, 3641, 3400, 3272, 3000, 2745, 2500, 2381, 2100, 1963, 1774, 1652, 1400, 1355, 1154, 1080, 950, 854, 762, 669, 640, 478, 438, 353, 300, 233, 150, 144, 100, 86, 60, 44, 34, 21, 15, 9, 3, 1, 1, 1, 1, 1, 1, 1, 1 };

const int AB_ID_timeMultByFreeFieldsEvenDepth[] =
	{ 200,  180,  160,  140,  120,  110,  105,  100,   98,   96,   96,   95,   94,   92,   91,   90,   89,   88,   86,   85,   83,   82,   81,   80,   78,   76,   74,   72,   70,   68,   66,   64,   62,   60,   55,   50,   45,   40,  33,  25,  18,  13,  11,  10,   9,   8,   7,   7,   6,   5,   5,   4,   4,   3,  3,  2,  2,  1,  1, 1, 1, 1};

const int AB_ID_timeMultByFreeFieldsOddDepth[] =
	{ 100,  100,  100,  100,   80,   60,   50,   40,   35,   30,   29,   28,   27,   26,   26,   26,   25,   24,   23,   22,   21,   20,   19,   18,   17,   16,   15,   14,   13,   12,   11,   10,   10,    9,    9,    8,    8,    7,   7,   6,   6,   5,   5,   4,   4,   3,   3,   3,   3,   2,   2,   2,   2,   2,  2,  2,  1,  1,  1, 1, 1, 1};

const int InitialStoneCounts[] = { CTZAARS, CTZARRAS, CTOTTS, 0, CTOTTS, CTZARRAS, CTZAARS };

// Position reprezentation
int board[BOARD_ARRAY_SIZE];
int stackHeights[BOARD_ARRAY_SIZE];
int player, moveNumber;

// History
int turnNumber;
Move *history[MAX_MOVES];

// Position properties useful for the evaluation function and the Move Ordering
int counts[STONE_TYPES];
int stoneSum;
int value;
int materialValue;
thash hash;
int highestStack[STONE_TYPES];
int highestStackPos[STONE_TYPES];
int highestStackPosCache[STONE_TYPES][MAX_STACK_HEIGHT];
int countsByHeight[STONE_TYPES][MAX_STACK_HEIGHT];
int zoneOfControl[STONE_TYPES], threatenByCounts[BOARD_ARRAY_SIZE];

// For saving
double searchDuration;

// For searching (AB and PNS)
int currDepth;	//for test
uns searchedNodes;

// Debug constants
#ifdef DEBUG
int moveAlive, entryAlive, ttHit, ttFound, ttKick, prunedCount, entry2Alive, tt2Kick, tt2Hit, tt2Found, ttCollision;
#endif

typedef int_fast64_t ttimestamp;

/// Returns current time in microseconds
static ttimestamp get_timer(void)
{
	struct timeval t;
	gettimeofday(&t, NULL);
	return 1000000LL * t.tv_sec + t.tv_usec;
}

/// Returns difference between tEnd time and tStart times in seconds
static double getDurationInSecs(ttimestamp tStart, ttimestamp tEnd)
{
	double res = (double) (tEnd - tStart) / 1e6;
	while (res < 0)
		res += 86400;
	return res;
}

/// Searches for the best move in a given position using algorithm determined by the parameter ai.
/// RecallAI is a parameter for DFPNS -- when it's not able to prove or disprove position, it calls Alpha-beta if recallAI is true.
/// Returns false when an error occurred, otherwise true.
int GetBestMove(int ai, int time, int depthAB, bool recallAI, Move **move1, Move **move2)
{
	ASSERT2(moveNumber == 1, "GetBest: moveNumber should be 1");
	if (moveNumber == 2) return false;
	*move1 = NULL;
	*move2 = NULL;
	if (ai == -1)
		ai = MAIN_AI;
	// switch between different search methods or AIs
	if (ai == AIALPHABETA) {	// Alpha-beta without iterative deepening (ID)
		searchedNodes = 0;
		// set debug counters
		DBG(moveAlive = entryAlive = ttHit = ttFound = ttKick = prunedCount = ttCollision = 0);
		int ret = 0;
		ttimestamp tStart = get_timer();
		DPRINT("ALPHA BETA WITH TT, stacks on board: %d", stoneSum);
		int depth = ALPHABETA_DEPTH;
		if (depthAB > -1) depth = depthAB;
		ret = AlphaBeta(depth, -WIN - 1, WIN + 1);
		ttimestamp tEnd = get_timer();
		searchDuration = getDurationInSecs(tStart, tEnd);
		value = ret;	// because of saving
		DPRINT("Alpha-Beta: time %0.3f s, searched: %d, value: %d, pruned %d", searchDuration, searchedNodes,
		       ret, prunedCount);
		DPRINT("Alive: move %d, entries %d, ttKicks %d, ttHits %d, ttFound %d", moveAlive,
		       entryAlive, ttKick, ttHit, ttFound);
		ASSERT(ttCollision == 0 || ttCollision > 1000000, "FOUND TT COLLISION: %d", ttCollision);
		TTEntry *saved = LookupPositionInTT();
		if (saved == NULL) {
			DPRINT("Error: cannot find position in TT!!!\n");
			return false;
		}
		*move1 = CloneMove(saved->bestMove1);
		if (saved->bestMove2 != NULL)
			*move2 = CloneMove(saved->bestMove2);
		else
			*move2 = NULL;
	} else if (ai >= AIALPHABETA_ID && ai <= AIALPHABETA_MAX) {	//Alpha-beta with ID
		int ret = 0;
		ttimestamp tStart = get_timer();
		ttimestamp tID;
		int depth = 2;
		double lastTime, currTime = 0;
		Move *m1 = NULL, *m2 = NULL, *lastm1, *lastm2;	//for better moves in losen positions (when using TT)
		int mult;
		do { // iterative deepening
			lastm1 = m1, lastm2 = m2;
			currDepth = depth;	// for getting branching factor on top level of the search
			searchedNodes = 0;
			// set debug counters
			DBG(moveAlive = entryAlive = ttHit = ttFound = ttKick = prunedCount = ttCollision = 0);
			if (ai == AIALPHABETA_ID) {	// alpha beta with TT and iterative deepening
				DPRINT("ALPHA BETA w. TT and ID, stacks on board: %d", stoneSum);
				ret = AlphaBeta(depth, -WIN, WIN);
			} else if (ai == AIALPHABETA_ID_PV) {	// alpha beta with TT, iterative deepening and move from TT
				DPRINT("ALPHA BETA WITH TT, ID and PV, stacks on board: %d", stoneSum);
				ret = AlphaBetaPV(depth, -WIN, WIN);
			} else if (ai == AIALPHABETA_ID_PV_MO) {
				DPRINT("ALPHA BETA w. TT, ID and PV and MO, stacks on board: %d", stoneSum);
				ret = AlphaBetaPVMO(depth, -WIN, WIN);
			} else if (ai == AIALPHABETA_ID_MO) {
				DPRINT("ALPHA BETA w. ID, MO, stacks on board: %d", stoneSum);
				ret = AlphaBetaMO(depth, -WIN, WIN, &m1, &m2);
			} else if (ai == AIALPHABETA_RANDOM) {	// alpha beta with random move selecting
				DPRINT("ALPHA BETA RANDOM w. TT, ID, PV and MO, stacks on board: %d, pl %d",
				       stoneSum, player);
				ret = AlphaBetaPVMORandom(depth, AI_RANDOM_MARGIN);
			} else if (ai == AIALPHABETA_ID_PV_MO_SCOUT) {
				DPRINT("ALPHA BETA w. TT, ID, PV, MO and NEGASCOUT, stacks on board: %d",
				       stoneSum);
				ret = AlphaBetaPVMONegascout(depth, -WIN, WIN);
			} else if (ai == AIALPHABETA_ID_PV_MO_HISTORY) {
				DPRINT("ALPHA BETA w. TT, ID, PV, MO and HISTORY, stacks on board: %d",
				       stoneSum);
				ret = AlphaBetaPVMOHistory(depth, -WIN, WIN);
			} else if (ai == AIALPHABETA_ID_PV_MO_SCOUT_HISTORY) {
				DPRINT("ALPHA BETA w. TT, ID, PV, MO, NEGASCOUT and HISTORY, stacks on board: %d",
				       stoneSum);
				ret = AlphaBetaPVMOHistoryNegascout(depth, -WIN, WIN);
			}
			DBG2(printZOCDebug());
			tID = get_timer();
			lastTime = currTime;
			currTime = getDurationInSecs(tStart, tID);
			DPRINT("Res: pl %d, dep. %d, dur. %0.3f s, searched %d, val. %d, pruned %d", player,
			       depth, currTime, searchedNodes, ret, prunedCount);
			DPRINT("Alive: move %d, entries %d, ttKicks %d, ttHits %d, ttFound %d",
			       moveAlive, entryAlive, ttKick, ttHit, ttFound);
			ASSERT(ttCollision == 0 || ttCollision > 1000000, "FOUND TT COLLISION: %d", ttCollision);
			if (ai != AIALPHABETA_ID_MO) {
				TTEntry *saved = LookupPositionInTT();
				if (saved == NULL) {
					DPRINT("Error: cannot find position in TT!!!\n");
					return false;
				}
				m1 = CloneMove(saved->bestMove1); // the clone is needed because of possible kicks from TT
				if (saved->bestMove2 != NULL)
					m2 = CloneMove(saved->bestMove2);
			}
			if (abs(ret) == WIN) {
				if (ret == -WIN && lastm1 != NULL && lastm2 != NULL) {
					DPRINT("AB: I am loser :(");
					m1 = lastm1, m2 = lastm2;
				}
				if (ret == WIN)
					DPRINT("AB: I am winner!!!");
				break;
			}
			// estimate time for the next depth -- it's mult * current duration
			mult = AB_ID_timeMultByFreeFieldsEvenDepth[TOTAL_STONES - stoneSum + depth];
			if (depth % 2 == 0)
				mult = AB_ID_timeMultByFreeFieldsOddDepth[TOTAL_STONES - stoneSum + depth];
			DPRINT("mult for next depth: %d", mult);
			depth += 1;
		} while (((depth <= MIN_AB_DEPTH && depthAB == -1)
			  || (depth <= depthAB && depthAB > -1)
			  || (currTime + (currTime - lastTime) * mult < time && depthAB == -1))
			 && abs(ret) < WIN);
		searchDuration = currTime;
		value = ret;	// because of saving
		*move1 = m1;
		if (stoneSum < 60 && m2 != NULL)
			*move2 = m2;
	} else if (ai >= DFPNS && ai <= DFPNS_MAX) { // depth-first proof-number search
		ASSERT(abs(value) != WIN, "dfpns cannot be called when someone won");
		searchedNodes = 0;
		ttimestamp tStart;
		tStart = get_timer();
		ttimestamp tID;
		maxDfpnsSearchedNodes = DFPNS_SEARCH_NODES;
		TT2Entry *saved = NULL;
		searchDuration = 0;
		FullMove *fm = NULL;
		do {
			// set debug counters
			DBG(moveAlive = entry2Alive = tt2Hit = tt2Found = tt2Kick = prunedCount = ttCollision = 0);
			if (ai == DFPNS) {
				DPRINT("DFPNS, stacks on board %d:", stoneSum);
				fm = dfpns(1, INFINITY_PNS, INFINITY_PNS);
			} else if (ai == DFPNS_EPS_TRICK) {
				DPRINT("DFPNS w. EPSILON TRICK, stacks on board %d:", stoneSum);
				fm = dfpnsEpsTrick(1, INFINITY_PNS, INFINITY_PNS);
			} else if (ai == WEAK_PNS) {
				DPRINT("DFPNS w. WEAK, stacks on board %d:", stoneSum);
				fm = weakpns(1, INFINITY_PNS, INFINITY_PNS);
			} else if (ai == DFPNS_EVAL_BASED) {
				DPRINT("DFPNS w. EVAL BASED, stacks on board %d:", stoneSum);
				fm = dfpnsEvalBased(1, INFINITY_PNS, INFINITY_PNS);
			} else if (ai == DFPNS_WEAK_EPS_EVAL) {
				DPRINT("DFPNS w. WEAK, EPS. TRICK, EVAL BASED INIT, stacks on board %d:", stoneSum);
				fm = dfpnsWeakEpsEval(1, INFINITY_PNS, INFINITY_PNS);
			} else if (ai == DFPNS_DYNAMIC_WIDENING_EPS_EVAL) {
				DPRINT("DFPNS w. DYNAMIC WIDENING, EPS. TRICK, EVAL BASED INIT, stacks on board %d:", stoneSum);
				fm = dfpnsDynWideningEpsEval(1, INFINITY_PNS, INFINITY_PNS);
			}
			tID = get_timer();
			searchDuration = (float) getDurationInSecs(tStart, tID);
			int ff = 0, tt = 0;
			if (fm != NULL) {	//fm NULL in not solved position
				ff = fm->m1->from, tt = fm->m1->to;
			}
			DPRINT("DFPNS: pl %d, duration %0.3f s, searched: %d", player, searchDuration, searchedNodes);
			if (fm != NULL) {
				ASSERT(fm->m1->from != fm->m1->to, "fm->m1->from %d, fm->m1->to %d SE ROVNA",
				       fm->m1->from, fm->m1->to);
				fm->m1->from = ff;
				fm->m1->to = tt;
			}
			DPRINT("Alive: move %d, entries %d, ttKicks %d, ttHits %d, ttFound %d",  moveAlive, entry2Alive, tt2Kick, tt2Hit, tt2Found);
			maxDfpnsSearchedNodes = (int) (((time - searchDuration) * searchedNodes * 1.1f) / searchDuration) + searchedNodes;	//* 1.1f because the estimation is too pesimistic
			if (maxDfpnsSearchedNodes - searchedNodes < 10000) maxDfpnsSearchedNodes = searchedNodes + 10000;
			DPRINT("next max dfpns searched nodes: %d", maxDfpnsSearchedNodes);
			saved = LookupPositionInTT2();
			if (saved == NULL) {
				DPRINT("Error: cannot find position in TT2!!!\n");
				return false;
			}
			DPRINT("DFPNS: pn = %d, dn = %d, searched %d", saved->pn, saved->dn, saved->searchedNodes);
		} while (searchDuration < time * 4 / 5.0f && saved->pn > 0 && saved->dn > 0 && saved->pn < INFINITY_PNS && saved->dn < INFINITY_PNS);	//AI_TIME_LIMIT * 4 / 5 -- because sometimes a few seconds are missing to time limit
		value = 0;
		if (saved->pn == 0 || saved->dn >= INFINITY_PNS) {
			ASSERT(fm != NULL, "fm in win pos NULL");
			DPRINT("saving move 1: from %s (%d), to %s (%d)", IndexToFieldName(fm->m1->from), fm->m1->from, IndexToFieldName(fm->m1->to), fm->m1->to);
			*move1 = fm->m1;
			*move2 = fm->m2;
			value = WIN;
			DPRINT("DFPNS: I am winner!!!");
		} else if (saved->dn == 0 || saved->pn >= INFINITY_PNS) {	//moves from TT are moves to pos with lost in highest depth
			ASSERT(fm != NULL, "fm in lost pos NULL");
			*move1 = fm->m1;
			*move2 = fm->m2;
			DPRINT("DFPNS: I'm loser :(");
			value = -WIN;
			if (recallAI) {
				int oldVal = value;
				value = 0;
				DPRINT("calling alpha-beta:");
				//call alpha-beta
				GetBestMove(AIALPHABETA_BEST, time, depthAB, false, move1, move2);
				DPRINT("end call alpha-beta");
				searchDuration += getDurationInSecs(tStart, tID);
				value = oldVal;
			}
		} else {
			value = 0;
			if (recallAI) {
				int oldVal = value;
				value = 0;
				DPRINT("calling alpha-beta:");
				//call alpha-beta
				GetBestMove(AIALPHABETA_BEST, time, depthAB, false, move1, move2);
				DPRINT("end call alpha-beta");
				searchDuration += getDurationInSecs(tStart, tID);
				value = oldVal;
			}
		}
	} else if (ai == BEGINNERS_AI) { // AI for beginners
		DPRINT("AI FOR BEGINNERS: AB PVMO beginner");
		FOR(i,0,15) { // set beginner material value constants
				StackHeightValue[i] = StackHeightValueBeginner[i];
		}
		int ret = 0;
		ttimestamp tStart = get_timer();
		ttimestamp tID;
		int depth = 2;
		double currTime = 0;
		Move *m1 = NULL, *m2 = NULL, *lastm1, *lastm2;	//for better moves in losen positions (when using TT)
		do {
			lastm1 = m1, lastm2 = m2;
			currDepth = depth;
			searchedNodes = 0;
			DBG(moveAlive = entryAlive = ttHit = ttFound = ttKick = prunedCount = ttCollision = 0);
			ret = AlphaBetaPVMORandomBeginner(depth, AI_RANDOM_MARGIN_BIGGER);
			DBG2(printZOCDebug());
			tID = get_timer();
			currTime = getDurationInSecs(tStart, tID);
			DPRINT("Alpha-Beta: pl %d, depth %d, time %0.3f s, searched: %d, value: %d, pruned %d", player,
			       depth, currTime, searchedNodes, ret, prunedCount);
			DPRINT("Alive: move %d, entries %d, ttKicks %d, ttHits %d, ttFound %d",
			       moveAlive, entryAlive, ttKick, ttHit, ttFound);
			ASSERT(ttCollision == 0 || ttCollision > 1000000, "FOUND TT COLLISION: %d", ttCollision);
			TTEntry *saved = LookupPositionInTT();
			if (saved == NULL) {
				DPRINT("Error: cannot find position in TT!!!\n");
				return false;
			}
			m1 = CloneMove(saved->bestMove1);
			if (saved->bestMove2 != NULL)
				m2 = CloneMove(saved->bestMove2);
			if (abs(ret) == WIN) {
				if (ret == -WIN && lastm1 != NULL && lastm2 != NULL) {
					DPRINT("AB: I am loser :(");
					m1 = lastm1, m2 = lastm2;
				}
				if (ret == WIN)
					DPRINT("AB: I am winner!!!");
				break;
			}
			depth += 1;
		} while (((depth <= BEGINNER_AB_DEPTH && depthAB == -1) || (depth <= depthAB && depthAB > -1)
			) && abs(ret) < WIN);
		searchDuration = currTime;
		value = ret;	// because of saving
		*move1 = m1;
		if (stoneSum < 60 && m2 != NULL)
			*move2 = m2;
	} else if (ai == INTERMEDIATE_AI) { // AI for intermediate players
		DPRINT("AI INTERMEDIATE: AB PVMO");
		int ret = 0;
		ttimestamp tStart = get_timer();
		ttimestamp tID;
		int depth = 2;
		double currTime = 0;
		Move *m1 = NULL, *m2 = NULL, *lastm1, *lastm2;	//for better moves in losen positions (when using TT)
		do {
			lastm1 = m1, lastm2 = m2;
			currDepth = depth;
			searchedNodes = 0;
			DBG(moveAlive = entryAlive = ttHit = ttFound = ttKick = prunedCount = ttCollision = 0);
			ret = AlphaBetaPVMORandom(depth, AI_RANDOM_MARGIN);
			DBG2(printZOCDebug());
			tID = get_timer();
			currTime = getDurationInSecs(tStart, tID);
			DPRINT("Alpha-Beta: pl %d, depth %d, time %0.3f s, searched: %d, value: %d, pruned %d", player,
			       depth, currTime, searchedNodes, ret, prunedCount);
			DPRINT("Alive: move %d, entries %d, ttKicks %d, ttHits %d, ttFound %d",
			       moveAlive, entryAlive, ttKick, ttHit, ttFound);
			ASSERT(ttCollision == 0 || ttCollision > 1000000, "FOUND TT COLLISION: %d", ttCollision);
			TTEntry *saved = LookupPositionInTT();
			if (saved == NULL) {
				DPRINT("Error: cannot find position in TT!!!\n");
				return false;
			}
			m1 = CloneMove(saved->bestMove1); // the clone is needed because of possible kicks from transposition table
			if (saved->bestMove2 != NULL)
				m2 = CloneMove(saved->bestMove2);
			if (abs(ret) == WIN) {
				if (ret == -WIN && lastm1 != NULL && lastm2 != NULL) {
					DPRINT("AB: I am loser :(");
					m1 = lastm1, m2 = lastm2;
				}
				if (ret == WIN)
					DPRINT("AB: I am winner!!!");
				break;
			}
			depth += 1;
		} while (((depth <= INTERMEDIATE_AB_DEPTH && depthAB == -1) || (depth <= depthAB && depthAB > -1)
			) && abs(ret) < WIN);
		searchDuration = currTime;
		value = ret;	//because of saving
		*move1 = m1;
		if (stoneSum < 60 && m2 != NULL)
			*move2 = m2;
	} else if (ai == AICOMBI_RANDOM_AB_PNS) { // the AI for experts
		DPRINT("COMBINATION OF AI: RANDOM, BEST ALPHA-BETA, BEST DFPNS");
		if (stoneSum > 55) {
			GetBestMove(AIALPHABETA_RANDOM, time, depthAB, recallAI, move1, move2);
		} else if (stoneSum > 23) {
			GetBestMove(AIALPHABETA_BEST, time, depthAB, recallAI, move1, move2);
		} else {
			GetBestMove(DFPNS_BEST, time, depthAB, recallAI, move1, move2);
		}
	} else {
		DPRINT("Error: Unknown AI!!!\n");
		return false;
	}
	ASSERT(turnNumber == 1 && moveNumber == 1,
		"Moves not reverted! turn number %d, moveNumber %d", turnNumber, moveNumber);
	if (searchDuration > time * 3) {
		DPRINT("HIGH SEARCH DURATION!");
	}
	return true;
}

