/*
 * WALTZ
 *
 * The module pns contains the implementation of the Depth-first Proof-number
 * Search (DFPNS) with some enhancements and the Transposition Table (TT) used
 * by DFPNS
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */

#include "ai/pns.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

uns maxDfpnsSearchedNodes;

static TT2Entry *DFPNSTranspositionTable[2 * TT2SIZE];	//2* because of replacement schema Twobig

ALWAYS_INLINE
TT2Entry *LookupPositionInTT2(void)
{
	uns index = hash % TT2SIZE;
	if (DFPNSTranspositionTable[index] != NULL && DFPNSTranspositionTable[index]->hash == hash)
		return DFPNSTranspositionTable[index];
	else if (DFPNSTranspositionTable[index + TT2SIZE] != NULL
		 && DFPNSTranspositionTable[index + TT2SIZE]->hash == hash)
		return DFPNSTranspositionTable[index + TT2SIZE];
	else
		return NULL;
}

ALWAYS_INLINE
static void FreeTT2Entry(TT2Entry * entry)
{
	ASSERT2(entry != NULL, "entry to free NULL");
	DBG(entry2Alive--);
	free(entry);
}

//TODO is the replacement scheme two big useful for dfpns?
ALWAYS_INLINE
static void AddPositionToTT2(uns pn, uns dn, uns minWinningDepth, uns maxLosingDepth, uns searchedNodes)
{
	uns index = hash % TT2SIZE;
	if (DFPNSTranspositionTable[index] != NULL && DFPNSTranspositionTable[index]->hash == hash) {
		if (searchedNodes > DFPNSTranspositionTable[index]->searchedNodes) {
			DFPNSTranspositionTable[index]->searchedNodes = searchedNodes;
			DFPNSTranspositionTable[index]->pn = pn;
			DFPNSTranspositionTable[index]->dn = dn;
			DFPNSTranspositionTable[index]->minWinningDepth = minWinningDepth;
			DFPNSTranspositionTable[index]->maxLosingDepth = maxLosingDepth;
		}
		return;
	}
	if (DFPNSTranspositionTable[index + TT2SIZE] != NULL && DFPNSTranspositionTable[index + TT2SIZE]->hash == hash) {
		if (searchedNodes > DFPNSTranspositionTable[index + TT2SIZE]->searchedNodes) {
			DFPNSTranspositionTable[index + TT2SIZE]->searchedNodes = searchedNodes;
			DFPNSTranspositionTable[index + TT2SIZE]->pn = pn;
			DFPNSTranspositionTable[index + TT2SIZE]->dn = dn;
			DFPNSTranspositionTable[index + TT2SIZE]->minWinningDepth = minWinningDepth;
			DFPNSTranspositionTable[index + TT2SIZE]->maxLosingDepth = maxLosingDepth;
		}
		return;
	}
	TT2Entry *entry = MALLOC(TT2Entry);
	DBG(entry2Alive++);
	entry->pn = pn;
	entry->dn = dn;
	entry->minWinningDepth = minWinningDepth;
	entry->maxLosingDepth = maxLosingDepth;
	entry->hash = hash;
	entry->searchedNodes = searchedNodes;
	TT2Entry *toSave = entry;
	if (DFPNSTranspositionTable[index] == NULL || DFPNSTranspositionTable[index]->searchedNodes < searchedNodes) {
		toSave = DFPNSTranspositionTable[index];
		DFPNSTranspositionTable[index] = entry;
	}
	if (toSave != NULL) {	// this holds: DFPNSTranspositionTable[index]->searchedNodes > DFPNSTranspositionTable[index + TT2SIZE]->searchedNodes
		TT2Entry *old = DFPNSTranspositionTable[index + TT2SIZE];
		DFPNSTranspositionTable[index + TT2SIZE] = toSave;
		if (old != NULL) {
			FreeTT2Entry(old);
			DBG(tt2Kick++);
			DPRINT2("tt2 kick");
		}
	}
}

//PN is for player to move
/// dfpns without enhancements
FullMove *dfpns(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	GenerateAllMovesSorted(&move);
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, sumDN = 0, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, sumDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, sumDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2();
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							pn = 1;
							dn = 1;
							winningDepth = INFINITY_PNS;
							losingDep = 3;	//3 is worst case (we count the maximum)
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(sumDN <= INFINITY_PNS && sumDN >= 0, "sumDN > INFINITY_PNS, sumDN = %d", sumDN);
					if (dn == INFINITY_PNS || sumDN == INFINITY_PNS)
						sumDN = INFINITY_PNS;
					else sumDN += dn; // not weak

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, sumDN %d", depth,
								minPN, sumDN);
							ASSERT(sumDN == INFINITY_PNS, "minPN == 0 and sumDN == %d", sumDN);
							if (depth == 1) {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (sumDN == 0 || sumDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			if (depth == 1 && sumDN == 0) {
				AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth, searchedNodes + 100); // fix bug, root position should be stored always
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		dfpns(depth + 1, tdn - sumDN + minDN, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

/// dfpns + 1 + Epsilon Trick
FullMove *dfpnsEpsTrick(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	GenerateAllMovesSorted(&move);
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, sumDN = 0, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, sumDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, sumDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2();
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							pn = 1;
							dn = 1;
							winningDepth = INFINITY_PNS;
							losingDep = 3;
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(sumDN <= INFINITY_PNS && sumDN >= 0, "sumDN > INFINITY_PNS, sumDN = %d", sumDN);
					if (dn == INFINITY_PNS || sumDN == INFINITY_PNS)
						sumDN = INFINITY_PNS;
					else sumDN += dn; // not weak

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, sumDN %d", depth,
								minPN, sumDN);
							ASSERT(sumDN == INFINITY_PNS, "minPN == 0 and sumDN == %d", sumDN);
							if (depth == 1) {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (sumDN == 0 || sumDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
			if (depth == 1 && sumDN == 0) {
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2 + minPN2 * DFPNS_EPS);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		dfpnsEpsTrick(depth + 1, tdn - sumDN + minDN, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

/// dfpns + Weak PNS
FullMove *weakpns(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	GenerateAllMovesSorted(&move);
	int currVal = StaticValue();
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, maxDN = 0, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, maxDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, maxDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2();
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							pn = 1;
							dn = 1;
							winningDepth = INFINITY_PNS;
							losingDep = 3;
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(maxDN <= INFINITY_PNS && maxDN >= 0, "maxDN > INFINITY_PNS, maxDN = %d", maxDN);
					if (dn == INFINITY_PNS || maxDN == INFINITY_PNS)
						maxDN = INFINITY_PNS;
					else
						maxDN = MAX(maxDN, dn);	//weak
					//else maxDN += dn; // not weak

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, maxDN %d", depth,
								minPN, maxDN);
							ASSERT(maxDN == INFINITY_PNS, "minPN == 0 and maxDN == %d", maxDN);
							if (depth == 1) {
								AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		if (maxDN > 0) {
			//maxDN += moveCount - 1;	//weak (usual)
			// weak with heuristic counting -- using a similar step function as eval based pns
			int step = 0;
			if (currVal >= -WPNS_T)
				step++;
			if (currVal >= WPNS_T)
				step++;
			maxDN += (moveCount - 1) * (step * WPNS_H);	// + 1
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (maxDN == 0 || maxDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
			if (depth == 1 && maxDN == 0) {
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		weakpns(depth + 1, tdn - maxDN + minDN, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

/// dfpns + Evaluation Function Based Enhancement
FullMove *dfpnsEvalBased(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	GenerateAllMovesSorted(&move);
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, sumDN = 0, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, sumDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, sumDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2();
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							//step function
							int step = 0;
							int val = StaticValue();
							if (val >= -EFBPNS_T)
								step++;
							if (val >= EFBPNS_T)
								step++;
							pn = ((2 - step) * EFBPNS_B + 1)
								* BranchingFactorByFreeFields[TOTAL_STONES - stoneSum];
							dn = 1 + EFBPNS_A * step;
							winningDepth = INFINITY_PNS;
							losingDep = 3;
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(sumDN <= INFINITY_PNS && sumDN >= 0, "sumDN > INFINITY_PNS, sumDN = %d", sumDN);
					if (dn == INFINITY_PNS || sumDN == INFINITY_PNS)
						sumDN = INFINITY_PNS;
					else sumDN += dn; // not weak

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, sumDN %d", depth,
								minPN, sumDN);
							ASSERT(sumDN == INFINITY_PNS, "minPN == 0 and sumDN == %d", sumDN);
							if (depth == 1) {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (sumDN == 0 || sumDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
			if (depth == 1 && sumDN == 0) {
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		dfpnsEvalBased(depth + 1, tdn - sumDN + minDN, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

/// dfpns + Weak PNS, 1 + Epsilon Trick, Evaluation Function Based Enhancement
FullMove *dfpnsWeakEpsEval(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	GenerateAllMovesSorted(&move);
	int currVal = StaticValue();
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, maxDN = 0, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, maxDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, maxDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2();
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							//step function
							int step = 0;
							int val = StaticValue();
							if (val >= -EFBPNS_T)
								step++;
							if (val >= EFBPNS_T)
								step++;
							pn = ((2 - step) * EFBPNS_B + 1)
								* BranchingFactorByFreeFields[TOTAL_STONES - stoneSum];
							dn = 1 + EFBPNS_A * step;
							winningDepth = INFINITY_PNS;
							losingDep = 3;
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(maxDN <= INFINITY_PNS && maxDN >= 0, "maxDN > INFINITY_PNS, maxDN = %d", maxDN);
					if (dn == INFINITY_PNS || maxDN == INFINITY_PNS)
						maxDN = INFINITY_PNS;
					else
						maxDN = MAX(maxDN, dn);	//weak
					//else maxDN += dn; // not weak

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, maxDN %d", depth,
								minPN, maxDN);
							ASSERT(maxDN == INFINITY_PNS, "minPN == 0 and maxDN == %d", maxDN);
							if (depth == 1) {
								AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		if (maxDN > 0) {
			//maxDN += moveCount - 1;	//weak (usual)
			// weak with heuristic counting -- using a similar step function as eval based pns
			int step = 0;
			if (currVal >= -WPNS_T)
				step = 1;
			if (currVal >= WPNS_T)
				step = 2;
			maxDN += (moveCount - 1) * (step * WPNS_H);	// + 1
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (maxDN == 0 || maxDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			AddPositionToTT2(minPN, maxDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
			if (depth == 1 && maxDN == 0) {
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2 + minPN2 * DFPNS_EPS);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		uns ntpn = tdn - maxDN + minDN;
		// check if thresholds are more than a constant
		//if (ntpn < 1000) ntpn = 1000;
		//if (ntdn < 1000) ntdn = 1000;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		dfpnsWeakEpsEval(depth + 1, ntpn, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

/// dfpns + Dynamic Widening, 1 + Epsilon Trick, Evaluation Function Based Enhancement
FullMove *dfpnsDynWideningEpsEval(uns depth, uns tpn, uns tdn)
{
	uns initSearchedNodes = searchedNodes;
	DPRINT2("PNS: player %d tpn %d tdn %d depth %d searched %u", player, tpn, tdn, depth, searchedNodes);
	ASSERT2(!IsEndOfGame(), "pns starting in a final position, val %d, depth %d", value, depth);
	Move *move, *move2, *curr, *curr2, *tmp;
	Move *maxLooseMove1 = NULL, *maxLooseMove2 = NULL;	// for counting the best move in lost position
	uns maxDNArray[DWPNS_J + 1], minPNArray[DWPNS_J + 1];
	GenerateAllMovesSorted(&move);
	while (1) {
		curr = move;
		Move *minPNm1 = NULL, *minPNm2 = NULL;
		uns minPN = INFINITY_PNS, minPN2 = INFINITY_PNS, minDN = INFINITY_PNS, moveCount = 0;
		uns minWinningDepth = INFINITY_PNS, maxLosingDepth = 0;	// for counting the best move in lost position
		FOR(i, 0, DWPNS_J) maxDNArray[i] = 0;
		FOR(i, 0, DWPNS_J) minPNArray[i] = INFINITY_PNS + 1;
		maxDNArray[DWPNS_J] = INFINITY_PNS + 1; // stopper
		minPNArray[DWPNS_J] = INFINITY_PNS + 1; // stopper
		while (curr != NULL) {
			ExecuteMove(curr);
			int plVal = player * value;
			if (plVal == WIN) {
				searchedNodes++;
				//imediate pruning
				RevertLastMove();
				DPRINT2("PNS: pruning fast after one move, depth %d, minPN 0, maxDN INFTY", depth);
				if (minPNm2 != NULL) {
					FreeMove(minPNm2);
				}
				AddPositionToTT2(0, INFINITY_PNS, 1, INFINITY_PNS, 1);	//searchedNodes - initSearchedNodes == 0
				if (depth == 1) {
					FreeAllMoves(move, curr);
					FullMove *fm = MALLOC(FullMove);
					fm->m1 = curr;
					fm->m2 = NULL;
					return fm;
				} else {
					FreeAllMovesWithoutException(move);
					return NULL;
				}
			}
			else {
				GenerateAllMovesSorted(&move2);
				curr2 = move2;
				while (curr2 != NULL) {
					ExecuteMove(curr2);
					searchedNodes++;
					uns pn, dn, winningDepth, losingDep;	// pn and dn are swaped between tree layers
					if (abs(value) == WIN) {
						if (player * value > 0) {
							pn = INFINITY_PNS;
							dn = 0;
							winningDepth = INFINITY_PNS;
							losingDep = 2;
						} else {
							//imediate pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning fast, depth %d, minPN 0, maxDN INFTY",
								depth);
							if (minPNm2 != NULL) {
								FreeMove(minPNm2);
							}
							AddPositionToTT2(0, INFINITY_PNS, 2, INFINITY_PNS, 2);	//searchedNodes - initSearchedNodes == 0; 2 is depth
							if (depth == 1) {
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
					} else {
						TT2Entry *entry2 = LookupPositionInTT2(); // TODO check for hash collisions?
						if (entry2 != NULL) {
							DBG(tt2Found++);
							pn = entry2->dn;
							dn = entry2->pn;
							winningDepth = entry2->maxLosingDepth + 2;
							losingDep = entry2->minWinningDepth + 2;
							if (winningDepth > INFINITY_PNS)
								winningDepth = INFINITY_PNS;
						} else {
							//step function
							int step = 0;
							int val = StaticValue();
							if (val >= -EFBPNS_T)
								step++;
							if (val >= EFBPNS_T)
								step++;
							pn = ((2 - step) * EFBPNS_B + 1)
								* BranchingFactorByFreeFields[TOTAL_STONES - stoneSum];
							dn = 1 + EFBPNS_A * step;
							winningDepth = INFINITY_PNS;
							losingDep = 3;	//3 is worst case (we count the maximum)
						}
					}
					moveCount++;

					if (winningDepth < minWinningDepth)
						minWinningDepth = winningDepth;
					if (losingDep > maxLosingDepth) {
						maxLosingDepth = losingDep;
						maxLooseMove1 = curr;
						maxLooseMove2 = CloneMove(curr2);	// otherwise it will be deallocated
					}

					ASSERT2(dn <= INFINITY_PNS && dn >= 0, "dn > INFINITY_PNS, dn = %d", dn);
					ASSERT2(maxDNArray[0] <= INFINITY_PNS && maxDNArray[0] >= 0, "maxDNArray[0] > INFINITY_PNS, maxDNArray[0] = %d", maxDNArray[0]);
					if (dn == INFINITY_PNS || maxDNArray[0] == INFINITY_PNS)
						maxDNArray[0] = INFINITY_PNS;
					else { // dynamic widening
						uns i = 0;
						/*DPRINT("adding pn %u dn %u, before:", pn, dn);
						FOR(k,0,DWPNS_J) printf("%u ", maxDNArray[k]);
						DPRINT(" ");
						FOR(k,0,DWPNS_J) printf("%u ", minPNArray[k]);
						DPRINT(", after:");*/
						while (pn >= minPNArray[i] && i < DWPNS_J)
							i++;
						if (i < DWPNS_J) {
							for (uns j = DWPNS_J - 1; j > i; j--) {
								maxDNArray[j] = maxDNArray[j - 1];
								minPNArray[j] = minPNArray[j - 1];
							}
							maxDNArray[i] = dn;
							minPNArray[i] = pn;
							/*FOR(k,0,DWPNS_J) printf("%u ", maxDNArray[k]);
							DPRINT(" ");
							FOR(k,0,DWPNS_J) printf("%u ", minPNArray[k]);
							DPRINT(" ");*/
						}
						//DPRINT(" ");
					}

					if (pn < minPN || (pn == minPN && dn < minDN)) {
						minPN2 = minPN;
						minPN = pn;
						minDN = dn;
						if (minPNm2 != NULL) {
							FreeMove(minPNm2);
						}
						if (minPN == 0) {	// pruning
							RevertLastMove();
							RevertLastMove();
							DPRINT2("PNS: pruning, depth %d, minPN %d, maxDN %d", depth,
								minPN, maxDNArray[0]);
							ASSERT(maxDNArray[0] == INFINITY_PNS, "minPN == 0 and maxDN == %d", maxDNArray[0]);
							if (depth == 1) {
								AddPositionToTT2(minPN, maxDNArray[0], minWinningDepth, maxLosingDepth,
										 searchedNodes + 100); // fix bug, root position should be stored always
								FreeAllMovesWithoutException(curr2->next);
								FreeAllMoves(move, curr);
								FullMove *fm = MALLOC(FullMove);
								fm->m1 = curr;
								fm->m2 = curr2;
								return fm;
							} else {
								AddPositionToTT2(minPN, maxDNArray[0], minWinningDepth, maxLosingDepth,
										 searchedNodes - initSearchedNodes);
								FreeAllMovesWithoutException(curr2);
								FreeAllMovesWithoutException(move);
								return NULL;
							}
						}
						minPNm1 = curr;
						minPNm2 = curr2;
					} else if (pn < minPN2)
						minPN2 = pn;
					RevertLastMove();
					tmp = curr2;
					curr2 = curr2->next;
					if (tmp != minPNm2)
						FreeMove(tmp);
				}
			}
			RevertLastMove();
			curr = curr->next;
		}
		// dynamic widening
		uns sumDN = 0;
		FOR(i, 0, DWPNS_J) {
		  sumDN += maxDNArray[i];
		}
		ASSERT2(minPN > 0, "dfpns: minPN should be > 0, but it's %d", minPN);
		if (sumDN == 0 || sumDN >= tdn || minPN >= tpn || searchedNodes > maxDfpnsSearchedNodes) {
			AddPositionToTT2(minPN, sumDN, minWinningDepth, maxLosingDepth, searchedNodes - initSearchedNodes);
			if (depth == 1 && sumDN == 0) {
				DPRINT("I'm loser and max losing depth is %d", maxLosingDepth);
				DPRINT("move 1: f %d t %d", maxLooseMove1->from, maxLooseMove1->to);
				DPRINT("move 2: f %d t %d", maxLooseMove2->from, maxLooseMove2->to);
				FreeAllMoves(move, maxLooseMove1);
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				FullMove *fm = MALLOC(FullMove);
				fm->m1 = maxLooseMove1;
				fm->m2 = maxLooseMove2;
				return fm;
			} else {
				if (minPNm2 != NULL)
					FreeMove(minPNm2);
				ASSERT2(minPN > 0, "exiting dfpns and minPN %d", minPN);
				FreeAllMovesWithoutException(move);
				return NULL;
			}
		}
		uns ntdn = MIN(tpn, 1 + minPN2 + minPN2 * DFPNS_EPS);
		if (minPN2 == INFINITY_PNS)
			ntdn = tpn;
		ExecuteMove(minPNm1);
		ExecuteMove(minPNm2);
		dfpnsDynWideningEpsEval(depth + 1, tdn - sumDN + minDN, ntdn);
		RevertLastMove();
		RevertLastMove();
		FreeMove(minPNm2);
	}
}

//TODO we do not need to execute the whole move before looking in TT for move 2, only hash is needed (whether the position is a win for a player must be checked after ExecutedMove)
