/*
 * WALTZ
 *
 * The module main parses command line arguments, loads the position and
 * starts the search.
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <getopt.h>
#include "ai/lib.h"
#include "ai/saveload.h"

static const char *options = "a:t:e:f:g:o:b:d:m:nh";
static const struct option long_options[] = {
	{"ai", 1, 0, 'a'},
	{"bestmove", 1, 0, 'b'},
	{"depth", 1, 0, 'd'},
	{"execute", 1, 0, 'e'},
	{"factor", 1, 0, 'f'},
	{"generate", 1, 0, 'g'},
	{"help", 1, 0, 'h'},
	{"moves", 1, 0, 'm'},
	{"no-alphabeta", 1, 0, 'n'},
	{"output", 1, 0, 'o'},
	{"timelimit", 1, 0, 't'},
	{0, 0, 0, 0}
};

static void printHelp(void) {
	printf("Searches for the best moves in a position in Tzaar: \n");
	printf("\t-a AI --ai=AI\t AI number (1-9, 20-25, 40-42)\n");
	printf("\t-d DEPTH --depth=DEPTH\t Depth of Alpha-Beta search (if not set, search duration estimation is used).\n");
	printf("\t-b FILE --bestmove=FILE\t Search for the best moves in a position stored in FILE. This is required option.\n");
	printf("\t-e FILE --execute=FILE\t Execute the best moves and then save the position to FILE.\n");
	printf("\t-f FILE --factor=FILE \t Get branching factor of the position given by -b option and save the result to FILE.\n");
	printf("\t-g FILE --generate=FILE \t Generate random starting position and save it to FILE.\n");
	printf("\t-m --moves=FILE\t Execute moves that are in FILE and save the resulting position into FILE.\n");
	printf("\t-n --no-alphabeta\t Do not call Alpha-Beta functions when DFPNS did not find a win.\n");
	printf("\t-o FILE --output=FILE\t File to save the best moves. By default the file specified by --bestmoves option is used.\n");
	printf("\t-t SECONDS --timelimit=SECONDS\t Set time limit of the search to SECONDS (default is %d).\n", AI_TIME_LIMIT);
}

static int ProcessPosition(int ai, int time, int depthAB, const char *fileWithPosition, const char *fileBestMoves, const char *fileEorExecutedPos, bool recallAI)
{
	if (fileWithPosition == NULL)
		return FILE_WITH_POSITION_NULL;
	if (fileBestMoves == NULL)
		return FILE_FOR_BEST_MOVES_NULL;
	int err;
	if ((err = LoadPosition(fileWithPosition)) != OK) {
		printf("Cannot load file '%s' to get best moves (err %d). Usage: ./tzaar -b [FILE]\n", fileWithPosition,
		       err);
		return CANNOT_LOAD_POSITION;
	}
	DPRINT2("selected ai: %d", ai);
	if (moveNumber == 2) {	// should not happen
		printf("Cannot get best moves for the second move of player's turn (ai %d).", ai);
		return BAD_STONE;
	}
	Move *m1, *m2;
	if (GetBestMove(ai, time, depthAB, recallAI, &m1, &m2)) {
		if ((err = SaveBestMoves(fileBestMoves, m1, m2, recallAI)) != OK) {
			printf("Cannot save best moves in file '%s' (err %d)\n", fileBestMoves, err);
			return CANNOT_SAVE_POSITION;
		}
		if (fileEorExecutedPos != NULL) {
			value = 0;	//because of tests
			ExecuteMove(m1);
			if (m2 != NULL) {
				ExecuteMove(m2);
			}
			else if (abs(value) != WIN) { // hack for the very first turn of a game
				Move *pass = MALLOC(Move);
				pass->from = -1; pass->to = -1;
				ExecuteMove(pass);
			}
			DPRINT2("Saving the executed position to '%s'", fileEorExecutedPos);	// nic nerekne o case :)
			SavePosition(fileEorExecutedPos);
		}
		return OK;
	} else {
		printf("Cannot get best moves (ai %d).", ai);
		return GET_BEST_MOVE_ERROR;
	}
}

int main(int argc, char *argv[])
{
	int ai = -1;
	int time = AI_TIME_LIMIT; // in seconds
	char *executeFile = NULL;
	char *fileWithPosition = NULL;
	char *fileBestMoves = NULL;
	char *fileMoves = NULL;
	int c, option_index;

	//get branching factor
	bool getBrFactor = false;
	char *factorFile = NULL;
	// do not call alpha-beta
	bool recallAI = true;
	// depth of Alpha-Beta search
	int depthAB = -1;

	while ((c = getopt_long(argc, argv, options, long_options, &option_index)) >= 0) {
		switch (c) {
		case 'a':
			sscanf(optarg, "%d", &ai);
			DPRINT2("Argument ai: %d", ai);
			break;
		case 'd':
			sscanf(optarg, "%d", &depthAB);
			DPRINT2("Set AB depth to %d", depthAB);
			break;
		case 'b':
			fileWithPosition = (char *) malloc(sizeof(char) * (strlen(optarg) + 5));
			strcpy(fileWithPosition, optarg);
			break;
		case 'e':
			executeFile = (char *) malloc(sizeof(char) * (strlen(optarg) + 5));
			strcpy(executeFile, optarg);
			DPRINT2("Position after executing best moves will be saved in '%s'", executeFile);
			break;
		case 'f':
			factorFile = (char *) malloc(sizeof(char) * (strlen(optarg) + 5));
			strcpy(factorFile, optarg);
			getBrFactor = true;
			break;
		case 'g':
			InitBoard(RANDOM);
			SavePosition(optarg);
			return 0;
		case 'h':
			printHelp();
			return 0;
		case 'm':
			fileMoves = (char *) malloc(sizeof(char) * (strlen(optarg) + 5));
			strcpy(fileMoves, optarg);
			break;
		case 'n':
			recallAI = false;
			break;
		case 'o':
			fileBestMoves = (char *) malloc(sizeof(char) * (strlen(optarg) + 5));
			strcpy(fileBestMoves, optarg);
			break;
		case 't':
			sscanf(optarg, "%d", &time);
			DPRINT2("argument time limit: %d", time);
			break;
		default:
			printf("Not known argument '%c'. Type tzaar -h for help.\n", c);
			break;
		}
	}
	if (fileWithPosition == NULL) {
		printf("File with a position was not specified. Printing usage:\n");
		printHelp();
		return 0;
	}
	if (fileBestMoves == NULL) {
		fileBestMoves = fileWithPosition;
	}
	int err;
	
	// execute moves and save the resulting position
	if (fileMoves != NULL) {
		if ((err = LoadPosition(fileWithPosition)) != OK) {
			printf("Cannot load file '%s' to get br. factor (err %d)\n", fileWithPosition, err);
			return 0;
		}
		Move *m1, *m2;
		if ((err = LoadMoves(fileMoves, &m1, &m2)) != OK) {
			printf("Cannot load file '%s' with moves (err %d)\n", fileMoves, err);
			return 0;
		}
		ExecuteMove(m1);
		if (m2 != NULL && abs(value) != WIN) {
			ExecuteMove(m2);
		}
		else if (abs(value) != WIN) {
			Move *pass = MALLOC(Move);
			pass->from = -1; pass->to = -1;
			ExecuteMove(pass);
		}
		SavePosition(fileMoves);
		return 0;
	}
	// get branching factor
	else if (getBrFactor) {
		if ((err = LoadPosition(fileWithPosition)) != OK) {
			printf("Cannot load file '%s' to get br. factor (err %d)\n", fileWithPosition, err);
			return 0;
		}
		if (IsEndOfGame()) return 0;
		Move *move, *move2;
		int cnt = 0, uniq = 0;
		GenerateAllMoves(&move);
		while (move != NULL) {
			ExecuteMove(move);
			int plVal = player * value;
			if (plVal != WIN) {
					GenerateAllMoves(&move2);
					while (move2 != NULL) {
						ExecuteMove(move2);
						cnt++;
						TTEntry *entry = LookupPositionInTT();
						if (entry == NULL) {
							AddPositionToTT(0, 0, 0, 0, NULL, NULL);
							uniq++;
						}
						RevertLastMove();
						move2 = move2->next;
					}
			}
			else {
				cnt++;
				uniq++;
			}
			RevertLastMove();
			move = move->next;
		}
		FILE *f = fopen(factorFile, "a+");
		if (f == NULL) {
			printf("Cannot save to file '%s'\n", factorFile);
			return ERROR;
		}
		fprintf(f, "%d %d %d\n", stoneSum, cnt, uniq);
		fclose(f);
		printf("%d %d %d\n", stoneSum, cnt, uniq);
		return 0;
	}
	else if ((err = ProcessPosition(ai, time, depthAB, fileWithPosition, fileBestMoves, executeFile, recallAI)) != OK) {
		printf("Processing position failed with error %d.\n", err);
	}
	return err;
}
